#if defined(__clang__)
#pragma once
#include "feely_pona_unity.h"
#endif

extern "C" __declspec(dllexport)
void TELY_DLL_Reload(void *user_data)
{
    TELY_Platform *platform = DQN_CAST(TELY_Platform *)user_data;
    Dqn_Library_SetPointer(platform->core.dqn_lib);
}

extern "C" __declspec(dllexport)
void TELY_DLL_Init(void *user_data)
{
    TELY_Platform *platform   = DQN_CAST(TELY_Platform *)user_data;
    TELY_DLL_Reload(user_data);

    {
        Dqn_Arena_TempMemoryScope(&platform->arena);
        TELY_ChunkPool pool = {};
        pool.arena = &platform->arena;

        void *bytes16 = TELY_ChunkPool_Alloc(&pool, 16);
        TELY_ChunkPool_Dealloc(&pool, bytes16);
        DQN_ASSERT(pool.slots[TELY_ChunkPoolSlotSize_16b]       == DQN_CAST(void *)(DQN_CAST(char *)bytes16 - sizeof(TELY_ChunkPoolSlot)));
        DQN_ASSERT(pool.slots[TELY_ChunkPoolSlotSize_16b]->next == nullptr);

        void *bytes17 = TELY_ChunkPool_Alloc(&pool, 17);
        TELY_ChunkPool_Dealloc(&pool, bytes17);
        DQN_ASSERT(pool.slots[TELY_ChunkPoolSlotSize_32b]       == DQN_CAST(void *)(DQN_CAST(char *)bytes17 - sizeof(TELY_ChunkPoolSlot)));
        DQN_ASSERT(pool.slots[TELY_ChunkPoolSlotSize_32b]->next == nullptr);

        void *bytes1 = TELY_ChunkPool_Alloc(&pool, 1);
        void *bytes2 = TELY_ChunkPool_Alloc(&pool, 1);
        TELY_ChunkPool_Dealloc(&pool, bytes1);
        TELY_ChunkPool_Dealloc(&pool, bytes2);
        DQN_ASSERT(pool.slots[TELY_ChunkPoolSlotSize_16b]       == DQN_CAST(void *)(DQN_CAST(char *)bytes2  - sizeof(TELY_ChunkPoolSlot)));
        DQN_ASSERT(pool.slots[TELY_ChunkPoolSlotSize_16b]->next == DQN_CAST(void *)(DQN_CAST(char *)bytes1  - sizeof(TELY_ChunkPoolSlot)));

        void *bytes128k = TELY_ChunkPool_Alloc(&pool, DQN_KILOBYTES(128));
        TELY_ChunkPool_Dealloc(&pool, bytes128k);
        DQN_ASSERT(pool.slots[TELY_ChunkPoolSlotSize_128k]       == DQN_CAST(void *)(DQN_CAST(char *)bytes128k - sizeof(TELY_ChunkPoolSlot)));
        DQN_ASSERT(pool.slots[TELY_ChunkPoolSlotSize_128k]->next == nullptr);
    }

    // NOTE: TELY Game =============================================================================

    TELY_Assets *assets    = &platform->assets;
    Feely_Pona *pona       = Dqn_Arena_New(&platform->arena, Feely_Pona, Dqn_ZeroMem_Yes);
    TELY_Game *game        = &pona->game;
    game->chunk_pool.arena = &platform->arena;
    platform->user_data    = game;
    {
        TELY_AssetSpriteSheet *sheet = &pona->hero_sprite_sheet;
        Dqn_ThreadScratch scratch    = Dqn_Thread_GetScratch(nullptr);
        Dqn_String8 sheet_path       = Dqn_FsPath_ConvertF(scratch.arena, "%.*s/adventurer-v1.5-sheet.png", DQN_STRING_FMT(assets->textures_dir));
        sheet->tex_handle            = platform->func_load_texture(assets, DQN_STRING8("Hero"), sheet_path);
        sheet->sprite_count          = 109;
        sheet->sprites_per_row       = 7;
        sheet->sprite_size           = Dqn_V2I_InitNx2(50, 37);

        TELY_AssetSpriteAnimation hero_anims[] = {
            {DQN_STRING8("Everything"),     /*index*/ 0,   /*count*/ sheet->sprite_count},
            {DQN_STRING8("Idle"),           /*index*/ 0,   /*count*/ 3},
            {DQN_STRING8("Run"),            /*index*/ 8,   /*count*/ 6},
            {DQN_STRING8("Jump"),           /*index*/ 14,  /*count*/ 10},
            {DQN_STRING8("Floor slide"),    /*index*/ 24,  /*count*/ 5},
            {DQN_STRING8("Unknown"),        /*index*/ 29,  /*count*/ 9},
            {DQN_STRING8("Attack A"),       /*index*/ 38,  /*count*/ 11},
            {DQN_STRING8("Attack B"),       /*index*/ 49,  /*count*/ 4},
            {DQN_STRING8("Attack C"),       /*index*/ 53,  /*count*/ 6},
            {DQN_STRING8("Hurt A"),         /*index*/ 59,  /*count*/ 5},
            {DQN_STRING8("Hurt B"),         /*index*/ 64,  /*count*/ 5},
            {DQN_STRING8("Unsheath sword"), /*index*/ 69,  /*count*/ 4},
            {DQN_STRING8("Sheath sword"),   /*index*/ 73,  /*count*/ 4},
            {DQN_STRING8("Air drift"),      /*index*/ 77,  /*count*/ 2},
            {DQN_STRING8("Air drop"),       /*index*/ 79,  /*count*/ 2},
            {DQN_STRING8("Ladder climb"),   /*index*/ 81,  /*count*/ 4},
            {DQN_STRING8("Chi push"),       /*index*/ 85,  /*count*/ 8},
            {DQN_STRING8("Leap slice A"),   /*index*/ 93,  /*count*/ 7},
            {DQN_STRING8("Leap slice B"),   /*index*/ 100, /*count*/ 3},
            {DQN_STRING8("Leap slice C"),   /*index*/ 103, /*count*/ 6},
        };

        pona->hero_sprite_anims = Dqn_Slice_Alloc<TELY_AssetSpriteAnimation>(&platform->arena, DQN_ARRAY_UCOUNT(hero_anims), Dqn_ZeroMem_No);
        DQN_MEMCPY(pona->hero_sprite_anims.data, &hero_anims, sizeof(hero_anims[0]) * DQN_ARRAY_UCOUNT(hero_anims));
    }

    game->entities    = Dqn_VArray_Init<TELY_GameEntity>(&platform->arena, 1024 * 8);
    game->root_entity = Dqn_VArray_Make(&game->entities, 1, Dqn_ZeroMem_No);
    Dqn_FArray_Add(&game->parent_entity_stack, game->root_entity->handle);

    // NOTE: Unit test DFS pre-order and post-order walk
    {
        // NOTE: Setup entity-tree =================================================================

        TELY_GameEntity *f = TELY_Game_MakeEntityPointerF(game, "F");
        TELY_Game_PushParentEntity(game, f->handle);
            TELY_GameEntity *b = TELY_Game_MakeEntityPointerF(game, "B");
            TELY_GameEntity *g = TELY_Game_MakeEntityPointerF(game, "G");
            TELY_Game_PushParentEntity(game, b->handle);
                TELY_GameEntity *a = TELY_Game_MakeEntityPointerF(game, "A");
                TELY_GameEntity *d = TELY_Game_MakeEntityPointerF(game, "D");
                TELY_Game_PushParentEntity(game, d->handle);
                    TELY_GameEntity *c = TELY_Game_MakeEntityPointerF(game, "C");
                    TELY_GameEntity *e = TELY_Game_MakeEntityPointerF(game, "E");
                TELY_Game_PopParentEntity(game);
            TELY_Game_PopParentEntity(game);

            TELY_Game_PushParentEntity(game, g->handle);
                TELY_GameEntity *i = TELY_Game_MakeEntityPointerF(game, "I");
                TELY_Game_PushParentEntity(game, i->handle);
                    TELY_GameEntity *h = TELY_Game_MakeEntityPointerF(game, "H");
                TELY_Game_PopParentEntity(game);
            TELY_Game_PopParentEntity(game);
        TELY_Game_PopParentEntity(game);

        // NOTE: Pre order test ====================================================================

        TELY_GameEntity *pre_order_walk[9] = {};
        Dqn_usize pre_order_walk_count     = 0;
        for (TELY_GameEntityIterator it = {}; TELY_Game_DFSPreOrderWalkEntityTree(game, &it, game->root_entity);) {
            DQN_ASSERT(pre_order_walk_count < DQN_ARRAY_UCOUNT(pre_order_walk));
            pre_order_walk[pre_order_walk_count++] = it.entity;
        }

        DQN_ASSERT(pre_order_walk_count == DQN_ARRAY_UCOUNT(pre_order_walk));
        DQN_ASSERT(pre_order_walk[0] == f);
        DQN_ASSERT(pre_order_walk[1] == b);
        DQN_ASSERT(pre_order_walk[2] == a);
        DQN_ASSERT(pre_order_walk[3] == d);
        DQN_ASSERT(pre_order_walk[4] == c);
        DQN_ASSERT(pre_order_walk[5] == e);
        DQN_ASSERT(pre_order_walk[6] == g);
        DQN_ASSERT(pre_order_walk[7] == i);
        DQN_ASSERT(pre_order_walk[8] == h);

        // NOTE: Post order test ===================================================================

        TELY_GameEntity *post_order_walk[9] = {};
        Dqn_usize post_order_walk_count                  = 0;
        for (TELY_GameEntityIterator it = {}; TELY_Game_DFSPostOrderWalkEntityTree(game, &it, game->root_entity);) {
            DQN_ASSERT(post_order_walk_count < DQN_ARRAY_UCOUNT(post_order_walk));
            post_order_walk[post_order_walk_count++] = it.entity;
        }

        DQN_ASSERT(post_order_walk_count == DQN_ARRAY_UCOUNT(post_order_walk));
        DQN_ASSERT(post_order_walk[0] == a);
        DQN_ASSERT(post_order_walk[1] == c);
        DQN_ASSERT(post_order_walk[2] == e);
        DQN_ASSERT(post_order_walk[3] == d);
        DQN_ASSERT(post_order_walk[4] == b);
        DQN_ASSERT(post_order_walk[5] == h);
        DQN_ASSERT(post_order_walk[6] == i);
        DQN_ASSERT(post_order_walk[7] == g);
        DQN_ASSERT(post_order_walk[8] == f);

        // NOTE: Cleanup ===========================================================================
        TELY_Game_DeleteEntity(game, game->root_entity->handle);
        DQN_ASSERT(game->root_entity->first_child == nullptr);
        DQN_ASSERT(game->root_entity->last_child  == nullptr);
        DQN_ASSERT(game->root_entity->next        == nullptr);
        DQN_ASSERT(game->root_entity->prev        == nullptr);
        DQN_ASSERT(game->root_entity->parent      == nullptr);
    }

    // NOTE: Hero
    {
        TELY_GameEntity *hero    = TELY_Game_MakeEntityPointerF(game, "Hero");
        hero->local_pos          = Dqn_V2_InitNx2(100.f, 100.f);
        hero->size_scale         = Dqn_V2_InitNx1(4);
        hero->sprite_sheet       = &pona->hero_sprite_sheet;
        hero->sprite_anims       = pona->hero_sprite_anims;
        hero->local_hit_box_size = Dqn_V2_InitV2I(pona->hero_sprite_sheet.sprite_size);
        hero->flags             |= TELY_EntityFlag_Clickable;
        hero->flags             |= TELY_EntityFlag_MoveByKeyboard;
        hero->flags             |= TELY_EntityFlag_MoveByMouse;
        game->clicked_entity     = hero->handle;
    }

    // NOTE: Enemy
    {
        TELY_GameEntity *enemy    = TELY_Game_MakeEntityPointerF(game, "Enemy");
        enemy->local_pos          = Dqn_V2_InitNx2(300.f, 300.f);
        enemy->size_scale         = Dqn_V2_InitNx1(4);
        enemy->sprite_sheet       = &pona->hero_sprite_sheet;
        enemy->sprite_anims       = pona->hero_sprite_anims;
        enemy->local_hit_box_size = Dqn_V2_InitV2I(pona->hero_sprite_sheet.sprite_size);
        enemy->flags             |= TELY_EntityFlag_Clickable;
        enemy->flags             |= TELY_EntityFlag_MoveByKeyboard;
        enemy->flags             |= TELY_EntityFlag_MoveByMouse;
    }

    // NOTE: Wall
    {
        TELY_GameEntity *entity    = TELY_Game_MakeEntityPointerF(game, "V. Wall");
        entity->local_pos          = Dqn_V2_InitNx2(100.f, 300.f);
        entity->local_hit_box_size = Dqn_V2_InitNx2(100.f, 300.f);
        entity->flags             |= TELY_EntityFlag_Clickable;
        entity->flags             |= TELY_EntityFlag_MoveByKeyboard;
        entity->flags             |= TELY_EntityFlag_MoveByMouse;

        TELY_GameShape *wall       = Dqn_FArray_Make(&entity->shapes, Dqn_ZeroMem_Yes);
        wall->type                 = TELY_GameShapeType_Rect;
        wall->p2                   = entity->local_hit_box_size;
        wall->colour               = TELY_COLOUR_GREEN_DARK_KHAKI_V4;
    }

    uint16_t font_size        = 18;
    game->camera.scale        = Dqn_V2_InitNx1(1);
    pona->inter_regular_font  = platform->func_load_font(assets,  DQN_STRING8("Inter (Regular)"),             DQN_STRING8("Data/Inter-Regular.otf"), font_size);
    pona->inter_italic_font   = platform->func_load_font(assets,  DQN_STRING8("Inter (Italic)"),              DQN_STRING8("Data/Inter-Italic.otf"),  font_size);
    pona->jetbrains_mono_font = platform->func_load_font(assets,  DQN_STRING8("JetBrains Mono NL (Regular)"), DQN_STRING8("Data/JetBrainsMonoNL-Regular.ttf"),  font_size);
    pona->test_audio          = platform->func_load_audio(assets, DQN_STRING8("Test Audio"),                  DQN_STRING8("Data/Audio/Purrple Cat - Moonwinds.qoa"));

    // NOTE: TELY audio ============================================================================

    TELY_Audio *audio = &platform->audio;
    audio->chunk_pool = &game->chunk_pool;

    // NOTE: TELY ui ===============================================================================

    TELY_UI *ui                           = &game->ui;
    ui->arena.allocs_are_allowed_to_leak = true;
}

void TELY_Game_EntityChangeState(TELY_GameEntity *entity, TELY_GameEntityState state)
{
    if (entity->state == state)
        return;

    entity->state      = state;
    entity->anim.frame = 0;
    entity->anim.ticks = 0;

    // decouple the state from the animation, e.g. the wall doesn't have an animation
    uint16_t desired_sprite_anim_index = 0;
    switch (state) {
        case TELY_GameEntityState_Nil: {
        } break;

        case TELY_GameEntityState_Idle: {
            desired_sprite_anim_index = TELY_Asset_GetSpriteAnim(entity->sprite_anims, DQN_STRING8("Idle"));
        } break;

        case TELY_GameEntityState_Attack: {
            desired_sprite_anim_index = TELY_Asset_GetSpriteAnim(entity->sprite_anims, DQN_STRING8("Attack A"));
        } break;

        case TELY_GameEntityState_Run: {
            desired_sprite_anim_index = TELY_Asset_GetSpriteAnim(entity->sprite_anims, DQN_STRING8("Running"));
        } break;
    }

    if (entity->sprite_sheet && entity->sprite_anims.size) {
        TELY_AssetSpriteAnimation const *sprite_anim = entity->sprite_anims.data + entity->anim.index;
        entity->anim.index                           = desired_sprite_anim_index;
    }
}

struct AStarNode
{
    Dqn_usize cost;
    Dqn_usize heuristic;
    Dqn_V2I   came_from;
};

Dqn_f32 const FP_TILE_SIZE = 37.f;
Dqn_Slice<Dqn_V2I> AStarPathFind(Dqn_Arena *arena, TELY_Platform *platform, Dqn_V2I src_tile, Dqn_V2I dest_tile)
{
    Dqn_DSMap<AStarNode> astar_info = Dqn_DSMap_Init<AStarNode>(128);
    DQN_DEFER { Dqn_DSMap_Deinit(&astar_info); };

    Dqn_FArray<Dqn_V2I, 128> frontier = {};
    Dqn_FArray_Add(&frontier, src_tile);

    Dqn_usize tile_count_x = DQN_CAST(Dqn_usize)(platform->core.window_size.w / FP_TILE_SIZE);
    Dqn_usize tile_count_y = DQN_CAST(Dqn_usize)(platform->core.window_size.h / FP_TILE_SIZE);

    // NOTE: Initialise the starting cost
    uint64_t src_tile_u64 = (DQN_CAST(uint64_t)src_tile.y << 32) | (DQN_CAST(uint64_t)src_tile.x << 0);
    Dqn_DSMap_MakeKeyU64(&astar_info, src_tile_u64);

    while (frontier.size) {
        Dqn_V2I curr_tile = Dqn_FArray_PopFront(&frontier, 1);
        if (curr_tile == dest_tile)
            break;

        Dqn_FArray<Dqn_V2I, 4> neighbours = {};
        {
            Dqn_V2I left   = Dqn_V2I_InitNx2(curr_tile.x - 1, curr_tile.y);
            Dqn_V2I right  = Dqn_V2I_InitNx2(curr_tile.x + 1, curr_tile.y);
            Dqn_V2I top    = Dqn_V2I_InitNx2(curr_tile.x,     curr_tile.y - 1);
            Dqn_V2I bottom = Dqn_V2I_InitNx2(curr_tile.x,     curr_tile.y + 1);

            if (left.x >= 0)
                Dqn_FArray_Add(&neighbours, left);
            if (right.x <= tile_count_x)
                Dqn_FArray_Add(&neighbours, right);
            if (top.y >= 0)
                Dqn_FArray_Add(&neighbours, top);
            if (bottom.y <= tile_count_y)
                Dqn_FArray_Add(&neighbours, bottom);
        }

        uint64_t const curr_tile_u64 = (DQN_CAST(uint64_t)curr_tile.y << 32) | (DQN_CAST(uint64_t)curr_tile.x << 0);
        Dqn_usize const curr_cost    = Dqn_DSMap_FindKeyU64(&astar_info, curr_tile_u64).value->cost;
        for (Dqn_V2I next_tile : neighbours) {

            Dqn_usize new_cost                          = curr_cost + 1;
            uint64_t next_tile_u64                      = (DQN_CAST(uint64_t)next_tile.y << 32) | (DQN_CAST(uint64_t)next_tile.x << 0);
            Dqn_DSMapResult<AStarNode> next_cost_result = Dqn_DSMap_MakeKeyU64(&astar_info, next_tile_u64);

            if (next_cost_result.found && new_cost >= next_cost_result.value->cost)
                continue;

            Dqn_usize manhattan_dist          = DQN_ABS(dest_tile.x - next_tile.x) + DQN_ABS(dest_tile.y - next_tile.y);
            next_cost_result.value->cost      = new_cost;
            next_cost_result.value->came_from = curr_tile;
            next_cost_result.value->heuristic = new_cost + manhattan_dist;

            // TODO(doyle): Find the insert location into the frontier
            bool inserted = false;
            DQN_FOR_UINDEX(index, frontier.size) {
                Dqn_V2I frontier_tile        = frontier.data[index];
                uint64_t frontier_tile_u64   = DQN_CAST(uint64_t)frontier_tile.y << 32 | DQN_CAST(uint64_t)frontier_tile.x << 0;
                Dqn_usize frontier_heuristic = Dqn_DSMap_FindKeyU64(&astar_info, frontier_tile_u64).value->heuristic;
                if (next_cost_result.value->heuristic >= frontier_heuristic)
                    continue;

                char const *src  = DQN_CAST(char *)(frontier.data + index);
                char const *dest = DQN_CAST(char *)(frontier.data + (index + 1));
                char const *end  = DQN_CAST(char *)(frontier.data + frontier.size);
                Dqn_usize bytes_to_move = end - src;
                DQN_MEMMOVE(DQN_CAST(void *)dest, src, bytes_to_move);

                frontier.data[index] = next_tile;
                frontier.size++;
                inserted = true;
                break;
            }

            if (inserted)
                continue;

            Dqn_FArray_Add(&frontier, next_tile);
        }
    }

    Dqn_usize slice_size = 0;
    for (Dqn_V2I it = dest_tile; it != src_tile; slice_size++) {
        uint64_t key_u64 = (DQN_CAST(uint64_t)it.y << 32) | (DQN_CAST(uint64_t)it.x << 0);
        it               = Dqn_DSMap_FindKeyU64(&astar_info, key_u64).value->came_from;
    }

    Dqn_Slice<Dqn_V2I> result = Dqn_Slice_Alloc<Dqn_V2I>(arena, slice_size, Dqn_ZeroMem_No);
    slice_size = 0;
    for (Dqn_V2I it = dest_tile; it != src_tile; ) {
        result.data[slice_size++] = it;
        uint64_t key_u64          = (DQN_CAST(uint64_t)it.y << 32) | (DQN_CAST(uint64_t)it.x << 0);
        it                        = Dqn_DSMap_FindKeyU64(&astar_info, key_u64).value->came_from;
    }

    DQN_ASSERT(result.size == slice_size);
    return result;
}

void FP_GameUpdate(TELY_Platform *platform, TELY_Game *game, TELY_Renderer *renderer, TELY_PlatformInput *input)
{
    if (TELY_Platform_InputKeyIsReleased(input->mouse_left))
        game->clicked_entity = game->prev_active_entity;

    Dqn_V2 dir_vector = {};
    if (TELY_Platform_InputScanCodeIsDown(input, TELY_PlatformInputScanCode_W))
        dir_vector.y = -1.f;
    if (TELY_Platform_InputScanCodeIsDown(input, TELY_PlatformInputScanCode_A))
        dir_vector.x = -1.f;
    if (TELY_Platform_InputScanCodeIsDown(input, TELY_PlatformInputScanCode_S))
        dir_vector.y = +1.f;
    if (TELY_Platform_InputScanCodeIsDown(input, TELY_PlatformInputScanCode_D))
        dir_vector.x = +1.f;

    if (game->clicked_entity.id) {
        if (TELY_Platform_InputScanCodeIsPressed(input, TELY_PlatformInputScanCode_Delete))
            TELY_Game_DeleteEntity(game, game->clicked_entity);

        TELY_GameEntity *player = TELY_Game_GetEntity(game, game->clicked_entity);
        if (player) {
            if (TELY_Platform_InputScanCodeIsDown(input, TELY_PlatformInputScanCode_J)) {
                TELY_Game_EntityChangeState(player, TELY_GameEntityState_Attack);
            }

            if (player->state != TELY_GameEntityState_Attack) {
                if (dir_vector.x || dir_vector.y) {
                    TELY_Game_EntityChangeState(player, TELY_GameEntityState_Run);
                } else {
                    TELY_Game_EntityChangeState(player, TELY_GameEntityState_Idle);
                }
            }
        }
    } else {
        game->camera.world_pos += dir_vector * 5.f;
    }

    for (TELY_GameEntityIterator it = {}; TELY_Game_DFSPostOrderWalkEntityTree(game, &it, game->root_entity); ) {
        TELY_GameEntity *entity  = it.entity;
        entity->alive_time_s    += input->delta_s;

        // NOTE: Move entity by keyboard ===========================================================
        Dqn_V2 acceleration = {};
        if (game->clicked_entity == entity->handle) {
            if (entity->flags & TELY_EntityFlag_MoveByKeyboard) {
                acceleration = dir_vector * 10000000.f;
                if (dir_vector.x)
                    entity->facing_left = dir_vector.x < 0.f;
            }
        }

        Dqn_V2 entity_world_pos = TELY_Game_CalcEntityWorldPos(game, entity->handle);
        if (entity->name == DQN_STRING8("Enemy")) {
            if (entity->handle != game->clicked_entity && entity->stalk_entity != game->clicked_entity) {
                entity->stalk_entity = game->clicked_entity;
            }

            TELY_GameEntity *stalk_entity = TELY_Game_GetEntity(game, entity->stalk_entity);
            if (stalk_entity) {
                Dqn_V2 stalk_world_pos = TELY_Game_CalcEntityWorldPos(game, stalk_entity->handle);
                Dqn_V2I stalk_tile     = Dqn_V2I_InitNx2(stalk_world_pos.x / FP_TILE_SIZE, stalk_world_pos.y / FP_TILE_SIZE);
                if (entity->stalk_entity_last_known_tile != stalk_tile) {
                    entity->stalk_entity_last_known_tile = stalk_tile;

                    // NOTE: Dealloc all waypoints
                    for (TELY_GameWaypoint *waypoint = entity->waypoints->next; waypoint != entity->waypoints; ) {
                        TELY_GameWaypoint *next = waypoint->next;
                        TELY_ChunkPool_Dealloc(&game->chunk_pool, waypoint);
                        waypoint               = next;
                    }
                    entity->waypoints->next = entity->waypoints;
                    entity->waypoints->prev = entity->waypoints;

                    Dqn_V2I entity_tile          = Dqn_V2I_InitNx2(entity_world_pos.x / FP_TILE_SIZE, entity_world_pos.y / FP_TILE_SIZE);
                    Dqn_Slice<Dqn_V2I> path_find = AStarPathFind(&platform->arena, platform, entity_tile, stalk_tile);

                    for (Dqn_usize index = path_find.size - 1; index < path_find.size; index--) {
                        TELY_GameWaypoint *waypoint = TELY_ChunkPool_New(&game->chunk_pool, TELY_GameWaypoint);
                        waypoint->pos               = path_find.data[index];
                        waypoint->next              = entity->waypoints;
                        waypoint->prev              = entity->waypoints->prev;
                        waypoint->next->prev        = waypoint;
                        waypoint->prev->next        = waypoint;
                    }
                }
            }
        }

        for (TELY_GameWaypoint *waypoint = entity->waypoints->next; waypoint != entity->waypoints; waypoint = waypoint->next) {
            Dqn_V2 circle_pos = Dqn_V2_InitNx2(waypoint->pos.x * FP_TILE_SIZE + FP_TILE_SIZE * .5f, waypoint->pos.y * FP_TILE_SIZE + FP_TILE_SIZE * .5f);
            TELY_Render_CircleColourV4(renderer, circle_pos, 4.f, TELY_RenderShapeMode_Fill, TELY_COLOUR_MAGENTA_V4);
        }

        if (entity->waypoints->next != entity->waypoints) {
            TELY_GameWaypoint *waypoint = entity->waypoints->next;
            Dqn_V2I target_tile         = entity->waypoints->next->pos;
            Dqn_V2 target_pos           = Dqn_V2_InitNx2(target_tile.x * FP_TILE_SIZE + FP_TILE_SIZE *.5f, target_tile.y * FP_TILE_SIZE + FP_TILE_SIZE * .5f);
            Dqn_V2 entity_to_target_pos = target_pos - entity_world_pos;

            if (Dqn_V2_LengthSq(entity_to_target_pos) < DQN_SQUARED(entity->local_hit_box_size.x * .5f)) {
                waypoint->next->prev = waypoint->prev;
                waypoint->prev->next = waypoint->next;
                TELY_ChunkPool_Dealloc(&game->chunk_pool, waypoint);
            } else {
                Dqn_V2 entity_to_target_pos_norm = Dqn_V2_Normalise(entity_to_target_pos);
                entity->local_pos               += entity_to_target_pos_norm * (entity->local_hit_box_size.x * .05f);
            }
        }

        // NOTE: Core equations of motion ==========================================================
        {
            // f"(t) = a
            // f'(t) = at + v
            // f (t) = 0.5f*a(t^2) + vt + p
            Dqn_f32 t          = DQN_CAST(Dqn_f32)DQN_SQUARED(input->delta_s);
            Dqn_f32 t_squared  = DQN_SQUARED(t);
            entity->velocity   = (acceleration * t) + entity->velocity * 0.82f;

            Dqn_V2 delta_p     = (acceleration * 0.5f * t_squared) + (entity->velocity * t);
            entity->local_pos += delta_p;

            Dqn_Rect world_hit_box      = TELY_Game_CalcEntityWorldHitBox(game, entity->handle);
            Dqn_Rect new_world_hit_box  = world_hit_box;
            new_world_hit_box.pos      += delta_p;

            for (TELY_GameEntityIterator collider_it = {}; TELY_Game_DFSPostOrderWalkEntityTree(game, &collider_it, game->root_entity); ) {
                TELY_GameEntity *collider = collider_it.entity;
                if (collider->handle == entity->handle)
                    continue;

                // TODO(doyle): Minkowski sweep?
                Dqn_Rect collider_world_hit_box = TELY_Game_CalcEntityWorldHitBox(game, collider->handle);
                if (Dqn_Rect_Intersects(new_world_hit_box, collider_world_hit_box)) {
                }
            }
        }

        // NOTE: Move entity by mouse ==============================================================
        if (game->active_entity == entity->handle && entity->flags & TELY_EntityFlag_MoveByMouse) {
            if (entity->flags & TELY_EntityFlag_MoveByMouse) {
                entity->velocity   = {};
                entity->local_pos += input->mouse_p_delta;
            }
        }

        if (entity->flags & TELY_EntityFlag_DeriveHitBoxFromChildrenBoundingBox) {
            Dqn_Rect children_bbox = {};

            // TODO(doyle): Is the hit box supposed to include the containing
            // entity itself? Not sure
            children_bbox.pos = TELY_Game_CalcEntityWorldPos(game, entity->handle);

            for (TELY_GameEntityIterator child_it = {}; TELY_Game_DFSPreOrderWalkEntityTree(game, &child_it, entity);) {
                TELY_GameEntity *child = child_it.entity;
                DQN_ASSERT(child != entity);

                Dqn_Rect bbox = TELY_Game_CalcEntityWorldBoundingBox(game, child->handle);
                children_bbox = Dqn_Rect_Union(children_bbox, bbox);
            }

            Dqn_Rect padded_bbox         = Dqn_Rect_Expand(children_bbox, 16.f);
            entity->local_hit_box_offset = padded_bbox.pos - entity->local_pos + (padded_bbox.size * .5f);
            entity->local_hit_box_size   = padded_bbox.size;
        }

        // NOTE: Tick entity action ================================================================
        {
            bool action_is_done = false;
            TELY_AssetSpriteAnimation const *sprite_anim = entity->sprite_anims.data + entity->anim.index;
            if (sprite_anim) {
                if (entity->anim.frame >= sprite_anim->count) { // NOTE: Animation is finished
                    entity->anim.frame = 0;
                    entity->anim.ticks = 0;
                    action_is_done     = true;
                } else {
                    if (entity->anim.ticks++ > 4 /*ticks_per_anim_frame*/) {
                        entity->anim.frame++;
                        entity->anim.ticks = 0;
                    }
                }
            }

            if (action_is_done) {
                switch (entity->state) {
                    case TELY_GameEntityState_Nil:  break;
                    case TELY_GameEntityState_Idle: break;
                    case TELY_GameEntityState_Attack: {
                        TELY_Game_EntityChangeState(entity, TELY_GameEntityState_Idle);
                    } break;

                    case TELY_GameEntityState_Run: {
                        if (dir_vector.x == 0 && dir_vector.y == 0)
                            TELY_Game_EntityChangeState(entity, TELY_GameEntityState_Idle);
                    } break;
                }
            }
        }

        // NOTE: Calculate entity attack box =======================================================
        if (entity->state == TELY_GameEntityState_Attack) {

            entity->attack_box_size                   = entity->local_hit_box_size;
            TELY_AssetSpriteSheet const *sprite_sheet = entity->sprite_sheet;
            if (sprite_sheet) {
                entity->attack_box_size = Dqn_V2_InitV2I(sprite_sheet->sprite_size);
            }

            // NOTE: Position the attack box
            if (entity->facing_left) {
                entity->attack_box_offset = Dqn_V2_InitNx2(entity->local_hit_box_offset.x - entity->attack_box_size.w,
                                                           entity->local_hit_box_offset.y);
            } else {
                entity->attack_box_offset = Dqn_V2_InitNx2(entity->local_hit_box_offset.x + entity->local_hit_box_size.w,
                                                           entity->local_hit_box_offset.y);
            }
        } else {
            entity->attack_box_size = {};
        }
    }

    // NOTE: Do attacks ============================================================================
    for (TELY_GameEntityIterator attacker_it = {}; TELY_Game_DFSPostOrderWalkEntityTree(game, &attacker_it, game->root_entity); ) {
        TELY_GameEntity *attacker = attacker_it.entity;

        // NOTE: Resolve attack boxes
        if (Dqn_V2_Area(attacker->attack_box_size)) {
            Dqn_Rect attacker_box = TELY_Game_CalcEntityAttackWorldHitBox(game, attacker->handle);
            for (TELY_GameEntityIterator defender_it = {}; TELY_Game_DFSPostOrderWalkEntityTree(game, &defender_it, game->root_entity); ) {
                TELY_GameEntity *defender = defender_it.entity;
                if (defender->handle == attacker->handle)
                    continue;

                Dqn_Rect defender_box = TELY_Game_CalcEntityWorldHitBox(game, defender->handle);
                Dqn_Rect hit_rect     = Dqn_Rect_Intersection(attacker_box, defender_box);

                if (!Dqn_Rect_Area(hit_rect))
                    continue;

                TELY_Render_CircleColourV4(renderer, hit_rect.pos, 10.f, TELY_RenderShapeMode_Fill, TELY_COLOUR_MAGENTA_V4);
                Dqn_V2 attacker_center_pos       = Dqn_Rect_Center(attacker_box);
                Dqn_V2 defender_center_pos       = Dqn_Rect_Center(defender_box);
                Dqn_V2 attacker_to_defender      = defender_center_pos - attacker_center_pos;
                Dqn_V2 attacker_to_defender_norm = Dqn_V2_Normalise(attacker_to_defender);

                TELY_Render_LineColourV4(renderer, defender_center_pos, defender_center_pos + (attacker_to_defender_norm * 100.f), TELY_COLOUR_RED_V4, 3.f);

                Dqn_V2 acceleration  = attacker_to_defender_norm * 10000000.f;
                Dqn_f32 t            = DQN_CAST(Dqn_f32)DQN_SQUARED(input->delta_s);
                Dqn_f32 t_squared    = DQN_SQUARED(t);

                Dqn_V2 delta_p       = (acceleration * 0.5f * t_squared) + (defender->velocity * t);
                defender->velocity   = (acceleration * t) + defender->velocity * 2.0f;
                defender->local_pos += delta_p;
            }
        }
    }
}

void FP_GameRender(TELY_Game *game, TELY_Platform *platform, TELY_Renderer *renderer)
{
    TELY_PlatformInput *input = &platform->input;
    Dqn_M2x3 model_view       = TELY_Game_CameraModelViewM2x3(game->camera, platform);
    Dqn_V2 world_mouse_p      = Dqn_M2x3_MulV2(model_view, input->mouse_p);

    // NOTE: Draw tiles ============================================================================
    Dqn_usize tile_count_x = DQN_CAST(Dqn_usize)(platform->core.window_size.w / FP_TILE_SIZE);
    Dqn_usize tile_count_y = DQN_CAST(Dqn_usize)(platform->core.window_size.h / FP_TILE_SIZE);

    for (Dqn_usize x = 0; x < tile_count_x; x++) {
        Dqn_V2 start = Dqn_V2_InitNx2((x + 1) * FP_TILE_SIZE, 0);
        Dqn_V2 end   = Dqn_V2_InitNx2(start.x, platform->core.window_size.h);
        TELY_Render_LineColourV4(renderer, start, end, TELY_Colour_V4Alpha(TELY_COLOUR_WHITE_V4, .25f), 1.f);
    }

    for (Dqn_usize y = 0; y < tile_count_y; y++) {
        Dqn_V2 start = Dqn_V2_InitNx2(0, (y + 1) * FP_TILE_SIZE);
        Dqn_V2 end   = Dqn_V2_InitNx2(platform->core.window_size.w, start.y);
        TELY_Render_LineColourV4(renderer, start, end, TELY_Colour_V4Alpha(TELY_COLOUR_WHITE_V4, .25f), 1.f);
    }

    // NOTE: Draw entities =========================================================================
    for (TELY_GameEntityIterator it = {}; TELY_Game_DFSPostOrderWalkEntityTree(game, &it, game->root_entity); ) {
        TELY_GameEntity *entity  = it.entity;
        entity->alive_time_s    += input->delta_s;

        // NOTE: Render shapes in entity ===========================================================
        Dqn_V2 world_pos = TELY_Game_CalcEntityWorldPos(game, entity->handle);
        for (TELY_GameShape const &shape_ : entity->shapes) {
            TELY_GameShape const *shape = &shape_;
            Dqn_V2 local_to_world_p1    = world_pos + shape->p1;
            Dqn_V2 local_to_world_p2    = world_pos + shape->p2;
            switch (shape->type) {
                case TELY_GameShapeType_None: {
                } break;

                case TELY_GameShapeType_Circle: {
                    TELY_Render_CircleColourV4(renderer, local_to_world_p1, shape->circle_radius, shape->render_mode, shape->colour);
                } break;

                case TELY_GameShapeType_Rect: {
                    Dqn_Rect rect = Dqn_Rect_InitV2x2(local_to_world_p1, local_to_world_p2 - local_to_world_p1);
                    rect.pos     -= rect.size * .5f;
                    TELY_Render_RectColourV4(renderer, rect, shape->render_mode, shape->colour);
                } break;

                case TELY_GameShapeType_Line: {
                    TELY_Render_LineColourV4(renderer, local_to_world_p1, local_to_world_p2, shape->colour, shape->line_thickness);
                } break;
            }
        }

        // NOTE: Render entity sprites =============================================================
        if (entity->sprite_sheet && entity->sprite_anims.size) {
            TELY_AssetSpriteSheet     const *sprite_sheet = entity->sprite_sheet;
            TELY_AssetSpriteAnimation const *sprite_anim  = entity->sprite_anims.data + entity->anim.index;

            Dqn_usize sprite_index        = (sprite_anim->index + (entity->anim.frame % sprite_anim->count)) % sprite_sheet->sprite_count;
            Dqn_usize sprite_sheet_row    = sprite_index / sprite_sheet->sprites_per_row;
            Dqn_usize sprite_sheet_column = sprite_index % sprite_sheet->sprites_per_row;

            Dqn_Rect src_rect = {};
            src_rect.pos.x    = DQN_CAST(Dqn_f32)(sprite_sheet_column * sprite_sheet->sprite_size.w);
            src_rect.pos.y    = DQN_CAST(Dqn_f32)(sprite_sheet_row    * sprite_sheet->sprite_size.y);
            src_rect.size.w   = DQN_CAST(Dqn_f32)sprite_sheet->sprite_size.w;
            src_rect.size.h   = DQN_CAST(Dqn_f32)sprite_sheet->sprite_size.h;

            Dqn_Rect dest_rect = {};
            dest_rect.size     = src_rect.size * entity->size_scale;
            dest_rect.pos      = world_pos - (dest_rect.size * .5f);

            if (entity->facing_left)
                dest_rect.size.w *= -1.f; // NOTE: Flip the texture horizontally

            TELY_Render_TextureColourV4(renderer, sprite_sheet->tex_handle, src_rect, dest_rect, TELY_COLOUR_WHITE_V4);
        }

        // NOTE: Render attack box =================================================================
        {
            Dqn_Rect attack_box = TELY_Game_CalcEntityAttackWorldHitBox(game, entity->handle);
            TELY_Render_RectColourV4(renderer, attack_box, TELY_RenderShapeMode_Line, TELY_COLOUR_RED_TOMATO_V4);
        }

        // NOTE: Render world position =============================================================
        TELY_Render_CircleColourV4(renderer, world_pos, 4.f, TELY_RenderShapeMode_Fill, TELY_COLOUR_RED_TOMATO_V4);

        // NOTE: Render hot/active entity ==========================================================
        Dqn_Rect world_hit_box = TELY_Game_CalcEntityWorldHitBox(game, entity->handle);
        if (game->clicked_entity == entity->handle) {
            TELY_Render_RectColourV4(renderer, world_hit_box, TELY_RenderShapeMode_Line, TELY_COLOUR_WHITE_PALE_GOLDENROD_V4);
        } else if (game->hot_entity == entity->handle || (entity->flags & TELY_EntityFlag_DrawHitBox)) {
            Dqn_V4 hot_colour = game->hot_entity == entity->handle ? TELY_COLOUR_RED_TOMATO_V4 : TELY_Colour_V4Alpha(TELY_COLOUR_YELLOW_SANDY_V4, .5f);
            TELY_Render_RectColourV4(renderer, world_hit_box, TELY_RenderShapeMode_Line, hot_colour);

        }

        if (game->clicked_entity == entity->handle || game->hot_entity == entity->handle) {
            if (entity->name.size) {
                Dqn_V2I player_tile       = Dqn_V2I_InitNx2(world_pos.x / FP_TILE_SIZE, world_pos.y / FP_TILE_SIZE);
                Dqn_V2 entity_world_pos   = TELY_Game_CalcEntityWorldPos(game, entity->handle);
                Dqn_ThreadScratch scratch = Dqn_Thread_GetScratch(nullptr);
                Dqn_String8 label         = Dqn_String8_InitF(scratch.allocator,
                                                              "%.*s (%.1f, %.1f) (%I32d, %I32d)",
                                                              DQN_STRING_FMT(entity->name),
                                                              entity_world_pos.x,
                                                              entity_world_pos.y,
                                                              player_tile.x,
                                                              player_tile.y);
                TELY_Render_Text(renderer, world_mouse_p, Dqn_V2_InitNx2(0.f, 1), label);
            }
        }
    }
}

extern "C" __declspec(dllexport)
void TELY_DLL_FrameUpdate(void *user_data)
{
    TELY_Platform      *platform = DQN_CAST(TELY_Platform *) user_data;
    TELY_PlatformInput *input    = &platform->input;
    TELY_Assets        *assets   = &platform->assets;
    TELY_Renderer      *renderer = &platform->renderer;
    Feely_Pona         *pona     = DQN_CAST(Feely_Pona *) platform->user_data;
    TELY_Game          *game     = &pona->game;
    TELY_UI            *ui       = &game->ui;

    TELY_UI_FrameSetup(ui, assets, &platform->frame_arena);
    TELY_UI_PushFont(ui, pona->jetbrains_mono_font);

    TELY_Render_ClearColourV3(renderer, TELY_COLOUR_BLACK_MIDNIGHT_V4.rgb);
    TELY_Render_PushFont(renderer, pona->jetbrains_mono_font);
    {
        Dqn_ThreadScratch scratch  = Dqn_Thread_GetScratch(nullptr);
        Dqn_String8Builder builder = {};
        builder.allocator          = scratch.allocator;

        Dqn_String8Builder_AppendF(&builder, "TELY");
        if (Dqn_String8_IsValid(platform->core.os_name))
            Dqn_String8Builder_AppendF(&builder, " | %.*s", DQN_STRING_FMT(platform->core.os_name));

        Dqn_String8Builder_AppendF(&builder,
                                   " | %dx%d %.1fHz | TSC %.1f GHz",
                                   platform->core.display.size.w,
                                   platform->core.display.size.h,
                                   platform->core.display.refresh_rate,
                                   platform->core.tsc_per_second / 1'000'000'000.0);

        if (platform->core.ram_mb)
            Dqn_String8Builder_AppendF(&builder, " | RAM %.1fGB", platform->core.ram_mb / 1024.0);

        Dqn_String8Builder_AppendF(&builder,
                                   " | Work %04.1fms/f (%04.1f%%) | %05.1f FPS | Frame %'I64u | Timer %.1fs",
                                   input->work_ms,
                                   input->work_ms * 100.0 / input->delta_ms,
                                   1000.0 / input->delta_ms,
                                   input->frame_counter,
                                   input->timer_s);


        Dqn_String8 info_label = Dqn_String8Builder_Build(&builder, scratch.allocator);
        TELY_Render_Text(renderer, /*position*/ Dqn_V2_InitNx1(10), /*align*/ Dqn_V2_InitNx1(0), info_label);
    }

    // =============================================================================================

    game->prev_clicked_entity = game->clicked_entity;
    game->prev_hot_entity     = game->hot_entity;
    game->prev_active_entity  = game->active_entity;
    game->hot_entity          = {};
    game->active_entity       = {};
    Dqn_FArray_Clear(&game->parent_entity_stack);
    Dqn_FArray_Add(&game->parent_entity_stack, game->root_entity->handle);

    Dqn_M2x3 model_view  = TELY_Game_CameraModelViewM2x3(game->camera, platform);
    Dqn_V2 world_mouse_p = Dqn_M2x3_MulV2(model_view, input->mouse_p);

    // =============================================================================================

    TELY_Audio *audio = &platform->audio;
    if (audio->playback_size == 0) {
        TELY_Audio_Play(audio, pona->test_audio, 1.f /*volume*/);
    }

    // =============================================================================================

    if (TELY_Platform_InputKeyWasDown(input->mouse_left) && TELY_Platform_InputKeyIsDown(input->mouse_left)) {
        if (game->prev_active_entity.id)
            game->active_entity = game->prev_active_entity;
    } else {
        for (TELY_GameEntityIterator it = {}; TELY_Game_DFSPreOrderWalkEntityTree(game, &it, game->root_entity); ) {
            TELY_GameEntity *entity = it.entity;
            if (entity->local_hit_box_size.x <= 0 || entity->local_hit_box_size.y <= 0)
                continue;

            if ((entity->flags & TELY_EntityFlag_Clickable) == 0)
                continue;

            Dqn_Rect world_hit_box = TELY_Game_CalcEntityWorldHitBox(game, entity->handle);
            if (!Dqn_Rect_ContainsPoint(world_hit_box, world_mouse_p))
                continue;

            game->hot_entity = entity->handle;
            if (TELY_Platform_InputKeyIsPressed(input->mouse_left)) {
                game->active_entity  = entity->handle;
                game->clicked_entity = entity->handle;
            }
        }
    }

    FP_GameUpdate(platform, game, renderer, input);
    FP_GameRender(game, platform, renderer);
    TELY_Audio_MixPlaybackSamples(audio, assets);
}
